"""
Handles calculations associated with PSC Gross Split.
"""
from dataclasses import dataclass, field
import numpy as np

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts import psc_tools
from pyscnomics.econ.selection import (FluidType, GrossSplitRegime, TaxRegime, TaxType, DeprMethod, OtherRevenue,
                                       InflationAppliedTo)
from pyscnomics.econ.results import CashFlow


class GrossSplitException(Exception):
    """Exception to raise for a misuse of GrossSplit class"""

    pass


class SunkCostException(Exception):
    """Exception to raise for a misuse of Sunk Cost Method"""

    pass


@dataclass
class GrossSplit(BaseProject):
    field_status: str = field(default='No POD')
    field_loc: str = field(default='Onshore')
    res_depth: str = field(default='<=2500')
    infra_avail: str = field(default='Well Developed')
    res_type: str = field(default='Conventional')
    api_oil: str = field(default='<25')
    domestic_use: str = field(default='50<=x<70')
    prod_stage: str = field(default='Secondary')
    co2_content: str = field(default='<5')
    h2s_content: str = field(default='<100')

    base_split_ctr_oil: float = field(default=0.43)
    base_split_ctr_gas: float = field(default=0.48)
    split_ministry_disc: float = field(default=0.08)

    oil_dmo_volume_portion: float = field(default=0.25)
    oil_dmo_fee_portion: float = field(default=1.0)
    oil_dmo_holiday_duration: int = field(default=60)

    gas_dmo_volume_portion: float = field(default=1.0)
    gas_dmo_fee_portion: float = field(default=1.0)
    gas_dmo_holiday_duration: int = field(default=60)

    conversion_bboe2bscf: float = field(default=5.6)

    _oil_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _gas_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _oil_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _gas_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _oil_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _gas_non_capital: np.ndarray = field(default=None, init=False, repr=False)

    _cumulative_prod: np.ndarray = field(default=None, init=False, repr=False)

    _oil_base_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_base_split: np.ndarray = field(default=None, init=False, repr=False)
    _variable_split: float = field(default=0, init=False, repr=False)
    _var_split_array: np.ndarray = field(default=None, init=False, repr=False)
    _oil_prog_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_prog_split: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_split: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_split: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_share_before_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_share_before_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _oil_gov_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_gov_share: np.ndarray = field(default=None, init=False, repr=False)
    _oil_total_expenses: np.ndarray = field(default=None, init=False, repr=False)
    _gas_total_expenses: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _gas_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _oil_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)

    _transfer_to_oil: np.ndarray = field(default=None, init=False, repr=False)
    _transfer_to_gas: np.ndarray = field(default=None, init=False, repr=False)
    _oil_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)
    _gas_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)

    _oil_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _gas_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)

    _oil_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _oil_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ddmo: np.ndarray = field(default=None, init=False, repr=False)

    _gas_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _gas_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ddmo: np.ndarray = field(default=None, init=False, repr=False)

    _oil_taxable_income: np.ndarray = field(default=None, init=False, repr=False)
    _gas_taxable_income: np.ndarray = field(default=None, init=False, repr=False)

    _tax_rate_arr: np.ndarray = field(default=None, init=False, repr=False)

    _oil_tax: np.ndarray = field(default=None, init=False, repr=False)
    _gas_tax: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)

    _oil_ctr_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_cashflow: np.ndarray = field(default=None, init=False, repr=False)

    _oil_government_take: np.ndarray = field(default=None, init=False, repr=False)
    _gas_government_take: np.ndarray = field(default=None, init=False, repr=False)

    _oil_cashflow: CashFlow = field(default=None, init=False, repr=False)
    _gas_cashflow: CashFlow = field(default=None, init=False, repr=False)

    # Consolidated Attributes
    _consolidated_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_tangible: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_intangible: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_opex: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_asr: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_non_capital: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_depreciation: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_undepreciated_asset: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_share_before_tf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_gov_share_before_tf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_total_expenses: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cost_tobe_deducted: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_carward_deduct_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_carward_cost_aftertf: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_share_after_transfer: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_dmo_volume: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_dmo_fee: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_taxable_income: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ctr_net_share: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_government_take: np.ndarray = field(default=None, init=False, repr=False)

    def _wrapper_variable_split(self,
                                regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019):

        if regime == GrossSplitRegime.PERMEN_ESDM_8_2017:
            variable_split_func = self._get_var_split_08_2017()

        elif (regime == GrossSplitRegime.PERMEN_ESDM_52_2017 or
              GrossSplitRegime.PERMEN_ESDM_20_2019 or
              GrossSplitRegime.PERMEN_ESDM_12_2020):
            variable_split_func = self._get_var_split_52_2017()

        else:
            variable_split_func = ValueError('Not Recognized Gross Split Regime')

        return variable_split_func

    def _get_var_split_08_2017(self):
        """
        A function to get the value of Variable Split based on the given parameters.

        Returns
        -------
        _variable_split: float
            The value of variable split.
        """
        params = {
            'field_status': {
                'POD I': 0.05,
                'POD II': 0,
                'POFD': 0,
                'No POD': -0.05,
            },
            'field_loc': {
                'Onshore': 0,
                'Offshore (0<h<=20)': 0.08,
                'Offshore (20<h<=50)': 0.1,
                'Offshore (50<h<=150)': 0.12,
                'Offshore (150<h<=1000)': 0.14,
                'Offshore (h>1000)': 0.16,
            },
            'res_depth': {
                '<=2500': 0,
                '>2500': 0.01,
            },
            'infra_avail': {
                'Well Developed': 0,
                'New Frontier': 0.02,
            },
            'res_type': {
                'Conventional': 0,
                'Non Conventional': 0.16,
            },
            'co2_content': {
                '<5': 0,
                '5<=x<10': 0.005,
                '10<=x<20': 0.01,
                '20<=x<40': 0.015,
                '40<=x<60': 0.02,
                'x>=60': 0.04,
            },
            'h2s_content': {
                '<100': 0,
                '100<=x<300': 0.005,
                '300<=x<500': 0.0075,
                'x>=500': 0.01,
            },
            'api_oil': {
                '<25': 0.01,
                '>=25': 0,
            },
            'domestic_use': {
                '<30': 0,
                '30<=x<50': 0.02,
                '50<=x<70': 0.03,
                '70<=x<100': 0.04,
            },
            'prod_stage': {
                'Primary': 0,
                'Secondary': 0.03,
                'Tertiary': 0.05,
            }
        }

        self._variable_split = sum(params[param][getattr(self, param)] for param in params)

    def _get_var_split_52_2017(self):
        """
        A function to get the value of Variable Split based on the given parameters.

        Returns
        -------
        _variable_split: float
            The value of variable split.
        """
        params = {
            'field_status': {
                'POD I': 0.05,
                'POD II': 0.03,
                'No POD': 0,
            },
            'field_loc': {
                'Onshore': 0,
                'Offshore (0<h<=20)': 0.08,
                'Offshore (20<h<=50)': 0.1,
                'Offshore (50<h<=150)': 0.12,
                'Offshore (150<h<=1000)': 0.14,
                'Offshore (h>1000)': 0.16,
            },
            'res_depth': {
                '<=2500': 0,
                '>2500': 0.01,
            },
            'infra_avail': {
                'Well Developed': 0,
                'New Frontier Offshore': 0.02,
                'New Frontier Onshore': 0.04,
            },
            'res_type': {
                'Conventional': 0,
                'Non Conventional': 0.16,
            },
            'co2_content': {
                '<5': 0,
                '5<=x<10': 0.005,
                '10<=x<20': 0.01,
                '20<=x<40': 0.015,
                '40<=x<60': 0.02,
                'x>=60': 0.04,
            },
            'h2s_content': {
                '<100': 0,
                '100<=x<1000': 0.01,
                '1000<=x<2000': 0.02,
                '2000<=x<3000': 0.03,
                '3000<=x<4000': 0.04,
                'x>=4000': 0.05,
            },
            'api_oil': {
                '<25': 0.01,
                '>=25': 0,
            },
            'domestic_use': {
                'x<30': 0,
                '30<=x<50': 0.02,
                '50<=x<70': 0.03,
                '70<=x<100': 0.04,
            },
            'prod_stage': {
                'Primary': 0,
                'Secondary': 0.06,
                'Tertiary': 0.1,
            }
        }

        self._variable_split = sum(params[param][getattr(self, param)] for param in params)

    def _wrapper_progressive_split(self,
                                   fluid: FluidType,
                                   price: float,
                                   cum: float,
                                   regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019):

        if regime == GrossSplitRegime.PERMEN_ESDM_20_2019:
            prog_split = self._get_prog_split_52_2017(fluid, price, cum)

        elif regime == GrossSplitRegime.PERMEN_ESDM_8_2017:
            prog_split = self._get_prog_split_08_2017(fluid, price, cum)

        else:
            prog_split = ValueError('Not Recognized Gross Split Regime')

        return prog_split

    @staticmethod
    def _get_prog_split_08_2017(fluid: FluidType, price: float, cum: float):
        # Indonesia's Ministry Regulations No.08 The Year of 2017. At Appendix B Progressive Component
        if fluid == FluidType.OIL:
            if price < 40:
                ps = 0.075
            elif 40 <= price < 55:
                ps = 0.05
            elif 55 <= price < 70:
                ps = 0.025
            elif 70 <= price < 85:
                ps = 0.0
            elif 85 <= price < 100:
                ps = -0.025
            elif 100 <= price < 115:
                ps = -0.05
            elif 115 <= price:
                ps = -0.075
            else:
                ps = 0
        else:
            ps = 0

        # Cumulative Progressive Split
        if 0 < cum < 1:
            px = 0.05
        elif 1 <= cum < 10:
            px = 0.04
        elif 10 <= cum < 20:
            px = 0.03
        elif 20 <= cum < 50:
            px = 0.02
        elif 50 <= cum < 150:
            px = 0.01
        elif 150 <= cum:
            px = 0
        else:
            raise ValueError('No Regulation exist regarding the cumulative value')

        ps = ps + px
        return ps

    @staticmethod
    def _get_prog_split_52_2017(fluid: FluidType, price: float, cum: float):
        # Indonesia's Ministry Regulations No.52 The Year of 2017. At Appendix B Progressive Component
        if fluid == FluidType.OIL:
            ps = np.where(price > 0,
                          (85 - price) * 0.25 / 100,
                          0)

        elif fluid == FluidType.GAS:
            if price < 7:
                ps = (7 - price) * 2.5 / 100
            elif price > 10:
                ps = (10 - price) * 2.5 / 100
            else:
                ps = 0
        else:
            raise ValueError('Unknown fluid type')

        # Cumulative Progressive Split
        if np.logical_and(np.greater_equal(cum, 0), np.less(cum, 30)):
            px = 0.1
        elif np.logical_and(np.greater_equal(cum, 30), np.less(cum, 60)):
            px = 0.09
        elif np.logical_and(np.greater_equal(cum, 60), np.less(cum, 90)):
            px = 0.08
        elif np.logical_and(np.greater_equal(cum, 90), np.less(cum, 125)):
            px = 0.06
        elif np.logical_and(np.greater_equal(cum, 125), np.less(cum, 175)):
            px = 0.04
        elif np.greater_equal(cum, 175):
            px = 0
        else:
            raise ValueError('No Regulation exist regarding the cumulative value')

        ps = ps + px
        return ps

    @staticmethod
    def _get_deductible_cost(ctr_gross_share, cost_tobe_deducted, carward_deduct_cost):

        carward_deduct_cost = np.concatenate((np.zeros(1), carward_deduct_cost[:-1]))

        return np.where(ctr_gross_share > (cost_tobe_deducted + carward_deduct_cost),
                        cost_tobe_deducted + carward_deduct_cost,
                        ctr_gross_share)

    def _get_sunk_cost(self, sunk_cost_reference_year: int):
        oil_cost_raw = (
                self._oil_tangible_expenditures
                + self._oil_non_capital
        )
        self._oil_sunk_cost = oil_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        gas_cost_raw = (
                self._gas_tangible_expenditures
                + self._gas_non_capital
        )
        self._gas_sunk_cost = gas_cost_raw[
                              : (sunk_cost_reference_year - self.start_date.year + 1)
                              ]

        if sunk_cost_reference_year == self.start_date.year:
            self._oil_sunk_cost = np.zeros(1)
            self._gas_sunk_cost = np.zeros(1)

    def run(self,
            sulfur_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            electricity_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_OIL_REVENUE,
            co2_revenue: OtherRevenue = OtherRevenue.ADDITION_TO_GAS_REVENUE,
            is_dmo_end_weighted=False,
            regime: GrossSplitRegime = GrossSplitRegime.PERMEN_ESDM_20_2019,
            tax_regime: TaxRegime = TaxRegime.NAILED_DOWN,
            tax_rate: float | np.ndarray = 0.22,
            sunk_cost_reference_year: int = None,
            depr_method: DeprMethod = DeprMethod.PSC_DB,
            decline_factor: float | int = 2,
            year_ref: int = None,
            tax_type: TaxType = TaxType.VAT,
            vat_rate: np.ndarray | float = 0.0,
            lbt_rate: np.ndarray | float = 0.0,
            inflation_rate: np.ndarray | float = 0.0,
            future_rate: float = 0.02,
            inflation_rate_applied_to: InflationAppliedTo = InflationAppliedTo.CAPEX
            ):

        # Configure Sunk Cost Reference Year
        if sunk_cost_reference_year is None:
            sunk_cost_reference_year = self.start_date.year

        if sunk_cost_reference_year > self.oil_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.oil_onstream_date}"
            )

        if sunk_cost_reference_year > self.gas_onstream_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the on stream date: {self.gas_onstream_date}"
            )

        if sunk_cost_reference_year < self.start_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is before the project start date: {self.start_date}"
            )

        if sunk_cost_reference_year > self.end_date.year:
            raise SunkCostException(
                f"Sunk Cost Reference Year {sunk_cost_reference_year} "
                f"is after the project end date: {self.end_date}"
            )

        # Configure year reference
        if year_ref is None:
            year_ref = self.start_date.year

        # Get the WAP Price
        self._get_wap_price()

        # Calculate expenditures for every cost components
        self._get_expenditures(
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
            future_rate=future_rate,
            inflation_rate_applied_to=inflation_rate_applied_to
        )

        # Get The Other Revenue as the chosen selection
        self._get_other_revenue(sulfur_revenue=sulfur_revenue,
                                electricity_revenue=electricity_revenue,
                                co2_revenue=co2_revenue)

        # Depreciation (tangible cost)
        (
            self._oil_depreciation,
            self._oil_undepreciated_asset,
        ) = self._oil_tangible.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        (
            self._gas_depreciation,
            self._gas_undepreciated_asset,
        ) = self._gas_tangible.total_depreciation_rate(
            depr_method=depr_method,
            decline_factor=decline_factor,
            year_ref=year_ref,
            tax_type=tax_type,
            vat_rate=vat_rate,
            lbt_rate=lbt_rate,
            inflation_rate=inflation_rate,
        )

        # Non Capital Cost
        self._oil_non_capital = (
                self._oil_intangible_expenditures
                + self._oil_opex_expenditures
                + self._oil_asr_expenditures
        )

        self._gas_non_capital = (
                self._gas_intangible_expenditures
                + self._gas_opex_expenditures
                + self._gas_asr_expenditures
        )

        # Get Sunk Cost
        self._get_sunk_cost(sunk_cost_reference_year)

        # Variable Split. -> Will set the value of _variable_split
        self._wrapper_variable_split(regime=regime)

        # Base Split
        self._oil_base_split = np.full_like(self.project_years, fill_value=self.base_split_ctr_oil, dtype=float)
        self._gas_base_split = np.full_like(self.project_years, fill_value=self.base_split_ctr_gas, dtype=float)

        # Variable Split
        self._var_split_array = np.full_like(self.project_years, fill_value=self._variable_split, dtype=float)

        # Cumulative Production
        # self._cumulative_prod = np.cumsum(self._oil_lifting.lifting_rate +
        #                                   (self._gas_lifting.lifting_rate / self.conversion_bboe2bscf))

        self._cumulative_prod = np.cumsum(np.divide(self._oil_lifting.get_lifting_rate_arr(), self._oil_lifting.get_lifting_ghv_arr(), where=self._oil_lifting.get_lifting_ghv_arr()!=0) +
                                          (np.divide(self._gas_lifting.get_lifting_rate_arr(), self._gas_lifting.get_lifting_ghv_arr(), where=self._gas_lifting.get_lifting_ghv_arr()!=0) / self.conversion_bboe2bscf))

        # Progressive Split
        vectorized_get_prog_split = np.vectorize(self._wrapper_progressive_split)

        self._oil_prog_split = vectorized_get_prog_split(
            fluid=self._oil_lifting.fluid_type,
            price=self._oil_lifting.get_price_arr(),
            cum=self._cumulative_prod,
            regime=regime
        )

        self._gas_prog_split = vectorized_get_prog_split(
            fluid=self._gas_lifting.fluid_type,
            price=self._gas_lifting.get_price_arr(),
            cum=self._cumulative_prod,
            regime=regime
        )

        # Ministerial Discretion
        minis_disc_array = np.full_like(self.project_years, fill_value=self.split_ministry_disc, dtype=float)

        # Total Contractor Split
        self._oil_ctr_split = (self._oil_base_split + self._var_split_array + self._oil_prog_split +
                               minis_disc_array)
        self._gas_ctr_split = (self._gas_base_split + self._var_split_array + self._gas_prog_split +
                               minis_disc_array)

        # Contractor Share
        self._oil_ctr_share_before_transfer = self._oil_revenue * self._oil_ctr_split
        self._gas_ctr_share_before_transfer = self._gas_revenue * self._gas_ctr_split

        # Government Share
        self._oil_gov_share = self._oil_revenue - self._oil_ctr_share_before_transfer
        self._gas_gov_share = self._gas_revenue - self._gas_ctr_share_before_transfer

        # Total Investment
        self._oil_total_expenses = (self._oil_tangible_expenditures + self._oil_intangible_expenditures +
                                    self._oil_opex_expenditures + self._oil_asr_expenditures)
        self._gas_total_expenses = (self._gas_tangible_expenditures + self._gas_intangible_expenditures +
                                    self._gas_opex_expenditures + self._gas_asr_expenditures)

        # Cost to be Deducted
        self._oil_cost_tobe_deducted = (self._oil_depreciation + self._oil_intangible_expenditures +
                                        self._oil_opex_expenditures + self._oil_asr_expenditures)
        self._gas_cost_tobe_deducted = (self._gas_depreciation + self._gas_intangible_expenditures +
                                        self._gas_opex_expenditures + self._gas_asr_expenditures)

        # Carry Forward Deductible Cost (In PSC Cost Recovery called Unrecovered Cost)
        self._oil_carward_deduct_cost = psc_tools.get_unrecovered_cost(depreciation=self._oil_depreciation,
                                                                       non_capital=self._oil_non_capital,
                                                                       revenue=self._oil_ctr_share_before_transfer,
                                                                       ftp_ctr=np.zeros_like(self.project_years),
                                                                       ftp_gov=np.zeros_like(self.project_years),
                                                                       ic=np.zeros_like(self.project_years))

        self._gas_carward_deduct_cost = psc_tools.get_unrecovered_cost(depreciation=self._gas_depreciation,
                                                                       non_capital=self._gas_non_capital,
                                                                       revenue=self._gas_ctr_share_before_transfer,
                                                                       ftp_ctr=np.zeros_like(self.project_years),
                                                                       ftp_gov=np.zeros_like(self.project_years),
                                                                       ic=np.zeros_like(self.project_years))

        # Deductible Cost (In PSC Cost Recovery called Cost Recovery)
        self._oil_deductible_cost = self._get_deductible_cost(ctr_gross_share=self._oil_ctr_share_before_transfer,
                                                              cost_tobe_deducted=self._oil_cost_tobe_deducted,
                                                              carward_deduct_cost=self._oil_carward_deduct_cost)

        self._gas_deductible_cost = self._get_deductible_cost(ctr_gross_share=self._gas_ctr_share_before_transfer,
                                                              cost_tobe_deducted=self._gas_cost_tobe_deducted,
                                                              carward_deduct_cost=self._gas_carward_deduct_cost)

        # Transfer
        self._transfer_to_oil, self._transfer_to_gas = psc_tools.get_transfer(
            oil_unrecovered=self._oil_carward_deduct_cost,
            gas_unrecovered=self._gas_carward_deduct_cost,
            oil_ets_pretransfer=self._oil_ctr_share_before_transfer,
            gas_ets_pretransfer=self._gas_ctr_share_before_transfer)

        # Carry Forward Deductible Cost After Transfer
        self._oil_carward_cost_aftertf = self._oil_carward_deduct_cost - self._transfer_to_gas
        self._gas_carward_cost_aftertf = self._gas_carward_deduct_cost - self._transfer_to_oil

        # Contractor Share After Transfer
        self._oil_ctr_share_after_transfer = (self._oil_ctr_share_before_transfer +
                                              self._transfer_to_oil -
                                              self._transfer_to_gas)

        self._gas_ctr_share_after_transfer = (self._gas_ctr_share_before_transfer +
                                              self._transfer_to_gas -
                                              self._transfer_to_oil)

        # Contractor Net Operating Profit
        self._oil_net_operating_profit = self._oil_ctr_share_after_transfer - self._oil_deductible_cost
        self._gas_net_operating_profit = self._gas_ctr_share_after_transfer - self._gas_deductible_cost

        # DMO
        self._oil_dmo_volume, self._oil_dmo_fee, self._oil_ddmo = psc_tools.get_dmo(
            onstream_date=self.oil_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.oil_dmo_holiday_duration,
            dmo_volume_portion=self.oil_dmo_volume_portion,
            dmo_fee_portion=self.oil_dmo_fee_portion,
            lifting=self._oil_lifting,
            price=self._oil_wap_price,
            ctr_pretax_share=1.0,
            unrecovered_cost=self._oil_carward_cost_aftertf,
            is_dmo_end_weighted=is_dmo_end_weighted)

        self._gas_dmo_volume, self._gas_dmo_fee, self._gas_ddmo = psc_tools.get_dmo(
            onstream_date=self.gas_onstream_date,
            start_date=self.start_date,
            project_years=self.project_years,
            dmo_holiday_duration=self.gas_dmo_holiday_duration,
            dmo_volume_portion=self.gas_dmo_volume_portion,
            dmo_fee_portion=self.gas_dmo_fee_portion,
            lifting=self._gas_lifting,
            price=self._gas_wap_price,
            ctr_pretax_share=1.0,
            unrecovered_cost=self._gas_carward_cost_aftertf,
            is_dmo_end_weighted=is_dmo_end_weighted)

        # Taxable Income
        self._oil_taxable_income = self._oil_net_operating_profit - self._oil_ddmo
        self._gas_taxable_income = self._gas_net_operating_profit - self._gas_ddmo

        # Tax Payment
        # Generating Tax array if tax_rate argument is a single value not array
        if isinstance(tax_rate, float):
            self._tax_rate_arr = np.full_like(self.project_years, tax_rate, dtype=float)

        # Generating Tax array based on the tax regime if tax_rate argument is None
        if tax_rate is None:
            self._tax_rate_arr = self._get_tax_by_regime(tax_regime=tax_regime)

        elif isinstance(tax_rate, np.ndarray):
            self._tax_rate_arr = tax_rate

        self._oil_tax = self._oil_taxable_income * self._tax_rate_arr
        self._gas_tax = self._gas_taxable_income * self._tax_rate_arr

        # Contractor Net Share
        self._oil_ctr_net_share = self._oil_taxable_income - self._oil_tax
        self._gas_ctr_net_share = self._gas_taxable_income - self._gas_tax

        # Contractor Cash Flow
        self._oil_ctr_cashflow = (self._oil_ctr_share_before_transfer - self._oil_total_expenses - self._oil_ddmo -
                                  self._oil_tax)
        self._gas_ctr_cashflow = (self._gas_ctr_share_before_transfer - self._gas_total_expenses - self._gas_ddmo -
                                  self._gas_tax)

        # Government Take
        self._oil_government_take = self._oil_gov_share + self._oil_ddmo + self._oil_tax
        self._gas_government_take = self._gas_gov_share + self._gas_ddmo + self._gas_tax

        # Consolidated attributes
        self._consolidated_revenue = self._oil_revenue + self._gas_revenue
        self._consolidated_tangible = self._oil_tangible_expenditures + self._gas_tangible_expenditures
        self._consolidated_intangible = self._oil_intangible_expenditures + self._gas_intangible_expenditures
        self._consolidated_sunk_cost = self._oil_sunk_cost + self._gas_sunk_cost
        self._consolidated_opex = self._oil_opex_expenditures + self._gas_opex_expenditures
        self._consolidated_asr = self._oil_asr_expenditures + self._gas_asr_expenditures
        self._consolidated_non_capital = self._oil_non_capital + self._gas_non_capital
        self._consolidated_depreciation = self._oil_depreciation + self._gas_depreciation
        self._consolidated_undepreciated_asset = self._oil_undepreciated_asset + self._gas_undepreciated_asset
        self._consolidated_ctr_share_before_tf = self._oil_ctr_share_before_transfer + self._gas_ctr_share_before_transfer
        self._consolidated_gov_share_before_tf = self._oil_gov_share + self._gas_gov_share
        self._consolidated_total_expenses = self._oil_total_expenses + self._gas_total_expenses
        self._consolidated_cost_tobe_deducted = self._oil_cost_tobe_deducted + self._gas_cost_tobe_deducted
        self._consolidated_carward_deduct_cost = self._oil_carward_deduct_cost + self._gas_carward_deduct_cost
        self._consolidated_deductible_cost = self._oil_deductible_cost + self._gas_deductible_cost
        self._consolidated_carward_cost_aftertf = self._oil_carward_cost_aftertf + self._gas_carward_cost_aftertf
        self._consolidated_ctr_share_after_transfer = self._oil_ctr_share_after_transfer + self._gas_ctr_share_after_transfer
        self._consolidated_net_operating_profit = self._oil_net_operating_profit + self._gas_net_operating_profit
        self._consolidated_dmo_volume = self._oil_dmo_volume + self._gas_dmo_volume
        self._consolidated_dmo_fee = self._oil_dmo_fee + self._gas_dmo_fee
        self._consolidated_ddmo = self._oil_ddmo + self._gas_ddmo
        self._consolidated_taxable_income = self._oil_taxable_income + self._gas_taxable_income
        self._consolidated_tax_payment = self._oil_tax + self._gas_tax
        self._consolidated_ctr_net_share = self._oil_ctr_net_share + self._gas_ctr_net_share
        self._consolidated_government_take = self._oil_government_take + self._gas_government_take
        self._consolidated_cashflow = self._oil_ctr_cashflow + self._gas_ctr_cashflow

    def __len__(self):
        return self.project_duration

    def __eq__(self, other):
        # Between two instances of Gross Split
        if isinstance(other, GrossSplit):
            return all(
                (
                    np.allclose(self._oil_lifting.lifting_rate, other._oil_lifting.lifting_rate),
                    np.allclose(self._gas_lifting.lifting_rate, other._gas_lifting.lifting_rate),
                    np.allclose(self._oil_revenue, other._oil_revenue),
                    np.allclose(self._gas_revenue, other._gas_revenue),
                    np.allclose(self._oil_tangible_expenditures, other._oil_tangible_expenditures),
                    np.allclose(self._gas_tangible_expenditures, other._gas_tangible_expenditures),
                    np.allclose(self._oil_intangible_expenditures, other._oil_intangible_expenditures),
                    np.allclose(self._gas_intangible_expenditures, other._gas_intangible_expenditures),
                    np.allclose(self._oil_opex_expenditures, other._oil_opex_expenditures),
                    np.allclose(self._gas_opex_expenditures, other._gas_opex_expenditures),
                    np.allclose(self._oil_asr_expenditures, other._oil_asr_expenditures),
                    np.allclose(self._gas_asr_expenditures, other._gas_asr_expenditures),
                )
            )

        else:
            return False

    def __lt__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() < other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __le__(self, other):
        if isinstance(other, GrossSplit):
            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() <= other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __gt__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() > other_total_base_cashflow.irr()

        else:
            raise GrossSplitException

    def __ge__(self, other):
        if isinstance(other, GrossSplit):

            self.run()
            other.run()

            self_total_base_cashflow = self._oil_base_cashflow + self._gas_base_cashflow
            other_total_base_cashflow = other._oil_base_cashflow + other._gas_base_cashflow

            return self_total_base_cashflow.irr() >= other_total_base_cashflow.irr()

        else:
            raise GrossSplitException



from dataclasses import dataclass, field
import numpy as np

from pyscnomics.econ.revenue import Lifting
from pyscnomics.econ.selection import FluidType
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts import psc_tools


@dataclass
class Transition:
    contract1: CostRecovery | GrossSplit
    contract2: CostRecovery | GrossSplit
    argument_contract1: dict
    argument_contract2: dict

    # Attributes to define later
    _contract1_transitioned: CostRecovery | GrossSplit = field(default=None, init=False, repr=False)
    _contract2_transitioned: CostRecovery | GrossSplit = field(default=None, init=False, repr=False)

    project_years: np.ndarray = field(default=None, init=False, repr=False)

    _oil_lifting: Lifting = field(default=None, init=False, repr=False)
    _gas_lifting: Lifting = field(default=None, init=False, repr=False)
    _sulfur_lifting: Lifting = field(default=None, init=False, repr=False)
    _electricity_lifting: Lifting = field(default=None, init=False, repr=False)
    _co2_lifting: Lifting = field(default=None, init=False, repr=False)

    _oil_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _gas_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _sulfur_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _electricity_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _co2_revenue: np.ndarray = field(default=None, init=False, repr=False)

    _oil_tangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_tangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_intangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_intangible_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_opex_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_opex_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _oil_asr_expenditures: np.ndarray = field(default=None, init=False, repr=False)
    _gas_asr_expenditures: np.ndarray = field(default=None, init=False, repr=False)

    _oil_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _gas_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _sulfur_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _electricity_wap_price: np.ndarray = field(default=None, init=False, repr=False)
    _co2_wap_price: np.ndarray = field(default=None, init=False, repr=False)

    _oil_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_deductible_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_unrec_cost: np.ndarray = field(default=None, init=False, repr=False)
    _gas_unrec_cost: np.ndarray = field(default=None, init=False, repr=False)
    _oil_ctr_ets: np.ndarray = field(default=None, init=False, repr=False)
    _gas_ctr_ets: np.ndarray = field(default=None, init=False, repr=False)
    _oil_gov_ets: np.ndarray = field(default=None, init=False, repr=False)
    _gas_gov_ets: np.ndarray = field(default=None, init=False, repr=False)
    _net_operating_profit: np.ndarray = field(default=None, init=False, repr=False)
    _ctr_net_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _ctr_ftp: np.ndarray = field(default=None, init=False, repr=False)
    _gov_ftp: np.ndarray = field(default=None, init=False, repr=False)
    _ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _government_take: np.ndarray = field(default=None, init=False, repr=False)

    # Consolidated Attributes
    _consolidated_revenue: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_sunk_cost: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_cashflow: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_ddmo: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_tax_payment: np.ndarray = field(default=None, init=False, repr=False)
    _consolidated_government_take: np.ndarray = field(default=None, init=False, repr=False)

    @staticmethod
    def _parse_dataclass(contract):
        # Getting the attributes of the contract
        attr_contract = vars(contract)

        # Condition where only one fluid is produced
        produced_fluid = [i.fluid_type for i in attr_contract['lifting']]

        if FluidType.OIL not in produced_fluid:
            attr_contract['oil_onstream_date'] = None
        if FluidType.GAS not in produced_fluid:
            attr_contract['gas_onstream_date'] = None

        if isinstance(contract, CostRecovery):
            new_contract = CostRecovery(
                start_date=attr_contract['start_date'],
                end_date=attr_contract['end_date'],
                oil_onstream_date=attr_contract['oil_onstream_date'],
                gas_onstream_date=attr_contract['gas_onstream_date'],
                lifting=attr_contract['lifting'],
                tangible_cost=attr_contract['tangible_cost'],
                intangible_cost=attr_contract['intangible_cost'],
                opex=attr_contract['opex'],
                asr_cost=attr_contract['asr_cost'],
                oil_ftp_is_available=attr_contract['oil_ftp_is_available'],
                oil_ftp_is_shared=attr_contract['oil_ftp_is_shared'],
                oil_ftp_portion=attr_contract['oil_ftp_portion'],
                gas_ftp_is_available=attr_contract['gas_ftp_is_available'],
                gas_ftp_is_shared=attr_contract['gas_ftp_is_shared'],
                gas_ftp_portion=attr_contract['gas_ftp_portion'],
                tax_split_type=attr_contract['tax_split_type'],
                condition_dict=attr_contract['condition_dict'],
                indicator_rc_icp_sliding=attr_contract['indicator_rc_icp_sliding'],
                oil_ctr_pretax_share=attr_contract['oil_ctr_pretax_share'],
                gas_ctr_pretax_share=attr_contract['gas_ctr_pretax_share'],
                oil_ic_rate=attr_contract['oil_ic_rate'],
                gas_ic_rate=attr_contract['gas_ic_rate'],
                ic_is_available=attr_contract['ic_is_available'],
                oil_cr_cap_rate=attr_contract['oil_cr_cap_rate'],
                gas_cr_cap_rate=attr_contract['gas_cr_cap_rate'],
                oil_dmo_volume_portion=attr_contract['oil_dmo_volume_portion'],
                oil_dmo_fee_portion=attr_contract['oil_dmo_fee_portion'],
                oil_dmo_holiday_duration=attr_contract['oil_dmo_holiday_duration'],
                gas_dmo_volume_portion=attr_contract['gas_dmo_volume_portion'],
                gas_dmo_fee_portion=attr_contract['gas_dmo_fee_portion'],
                gas_dmo_holiday_duration=attr_contract['gas_dmo_holiday_duration'])
        else:
            new_contract = GrossSplit(
                start_date=attr_contract['start_date'],
                end_date=attr_contract['end_date'],
                oil_onstream_date=attr_contract['oil_onstream_date'],
                gas_onstream_date=attr_contract['gas_onstream_date'],
                lifting=attr_contract['lifting'],
                tangible_cost=attr_contract['tangible_cost'],
                intangible_cost=attr_contract['intangible_cost'],
                opex=attr_contract['opex'],
                asr_cost=attr_contract['asr_cost'],
                field_status=attr_contract['field_status'],
                field_loc=attr_contract['field_loc'],
                res_depth=attr_contract['res_depth'],
                infra_avail=attr_contract['infra_avail'],
                res_type=attr_contract['res_type'],
                api_oil=attr_contract['api_oil'],
                domestic_use=attr_contract['domestic_use'],
                prod_stage=attr_contract['prod_stage'],
                co2_content=attr_contract['co2_content'],
                h2s_content=attr_contract['h2s_content'],
                base_split_ctr_oil=attr_contract['base_split_ctr_oil'],
                base_split_ctr_gas=attr_contract['base_split_ctr_gas'],
                split_ministry_disc=attr_contract['split_ministry_disc'],
                oil_dmo_volume_portion=attr_contract['oil_dmo_volume_portion'],
                oil_dmo_fee_portion=attr_contract['oil_dmo_fee_portion'],
                oil_dmo_holiday_duration=attr_contract['oil_dmo_holiday_duration'],
                gas_dmo_volume_portion=attr_contract['gas_dmo_volume_portion'],
                gas_dmo_fee_portion=attr_contract['gas_dmo_fee_portion'],
                gas_dmo_holiday_duration=attr_contract['gas_dmo_holiday_duration'],
                conversion_bboe2bscf=attr_contract['conversion_bboe2bscf'])

        return new_contract

    def run(self, unrec_portion: float = 0.0):
        # Defining the transition start date and end date
        start_date_trans = min([self.contract1.start_date, self.contract2.start_date])
        end_date_trans = max([self.contract1.end_date, self.contract2.end_date])

        # Condition where the first contract end on the 31 December and new contract start on  1 January
        condition = np.logical_and(self.contract1.end_date.day == 31, self.contract1.end_date.month == 12)
        if condition:
            additional_year = 1
        else:
            additional_year = 0

        project_years_trans = np.arange(start_date_trans.year, end_date_trans.year + 1 + additional_year)
        project_duration_trans = end_date_trans.year - start_date_trans.year + 1 + additional_year

        # Defining the row gap and description between the prior contract and the new contract
        zeros_to_new = np.zeros(self.contract2.project_duration - 1 + additional_year)
        zeros_to_prior = np.zeros(self.contract1.project_duration - 1 + additional_year)
        desc_to_new = ['-'] * (self.contract2.project_duration - 1 + additional_year)
        desc_to_prior = ['-'] * (self.contract1.project_duration - 1 + additional_year)
        fluid_to_new = [FluidType.OIL] * (self.contract2.project_duration - 1 + additional_year)
        fluid_to_prior = [FluidType.OIL] * (self.contract1.project_duration - 1 + additional_year)
        ic_to_new = [False] * (self.contract2.project_duration - 1 + additional_year)
        ic_to_prior = [False] * (self.contract1.project_duration - 1 + additional_year)

        # Defining the array of years between the prior contract to the new contract
        years_to_new = np.arange(self.contract1.end_date.year + 1, self.contract2.end_date.year + 1 + additional_year)
        years_to_prior = np.arange(self.contract1.start_date.year, self.contract2.start_date.year + additional_year)

        # Modifying the contract1
        # Changing the start_date and end_date of contract1
        self.contract1.start_date = start_date_trans
        self.contract1.end_date = end_date_trans
        self.contract1.project_duration = project_duration_trans
        self.contract1.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract1 lifting
        for lift in self.contract1.lifting:
            lift.start_year = start_date_trans.year
            lift.end_year = end_date_trans.year

            lift.lifting_rate = np.concatenate((lift.lifting_rate, zeros_to_new))
            lift.price = np.concatenate((lift.price, zeros_to_new))
            lift.ghv = np.concatenate((lift.ghv, zeros_to_new))
            lift.prod_rate = np.concatenate((lift.prod_rate, zeros_to_new))
            lift.prod_year = np.concatenate((lift.prod_year, years_to_new))

            lift.project_duration = project_duration_trans
            lift.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract1 tangible
        for tang in self.contract1.tangible_cost:
            tang.start_year = start_date_trans.year
            tang.end_year = end_date_trans.year

            tang.cost = np.concatenate((tang.cost, zeros_to_new))
            tang.expense_year = np.concatenate((tang.expense_year, years_to_new)).astype(int)
            tang.cost_allocation = tang.cost_allocation + fluid_to_new
            tang.description = tang.description + desc_to_new
            tang.vat_portion = np.concatenate((tang.vat_portion, zeros_to_new))
            tang.vat_discount = np.concatenate((tang.vat_discount, zeros_to_new))
            tang.lbt_portion = np.concatenate((tang.lbt_portion, zeros_to_new))
            tang.lbt_discount = np.concatenate((tang.lbt_discount, zeros_to_new))
            tang.pis_year = np.concatenate((tang.pis_year, zeros_to_new)).astype(int)
            tang.salvage_value = np.concatenate((tang.salvage_value, zeros_to_new))
            tang.useful_life = np.concatenate((tang.useful_life, zeros_to_new))
            tang.depreciation_factor = np.concatenate((tang.depreciation_factor, zeros_to_new))
            tang.is_ic_applied = tang.is_ic_applied + ic_to_new

            tang.project_duration = project_duration_trans
            tang.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract1 intangible
        for intang in self.contract1.intangible_cost:
            intang.start_year = start_date_trans.year
            intang.end_year = end_date_trans.year

            intang.cost = np.concatenate((intang.cost, zeros_to_new))
            intang.expense_year = np.concatenate((intang.expense_year, years_to_new)).astype(int)
            intang.cost_allocation = intang.cost_allocation + fluid_to_new
            intang.description = intang.description + desc_to_new
            intang.vat_portion = np.concatenate((intang.vat_portion, zeros_to_new))
            intang.vat_discount = np.concatenate((intang.vat_discount, zeros_to_new))
            intang.lbt_portion = np.concatenate((intang.lbt_portion, zeros_to_new))
            intang.lbt_discount = np.concatenate((intang.lbt_discount, zeros_to_new))

            intang.project_duration = project_duration_trans
            intang.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract1 opex
        for opx in self.contract1.opex:
            opx.start_year = start_date_trans.year
            opx.end_year = end_date_trans.year

            opx.expense_year = np.concatenate((opx.expense_year, years_to_new)).astype(int)
            opx.cost_allocation = opx.cost_allocation + fluid_to_new
            opx.description = opx.description + desc_to_new
            opx.fixed_cost = np.concatenate((opx.fixed_cost, zeros_to_new))
            opx.prod_rate = np.concatenate((opx.prod_rate, zeros_to_new))
            opx.cost_per_volume = np.concatenate((opx.cost_per_volume, zeros_to_new))
            opx.variable_cost = np.concatenate((opx.variable_cost, zeros_to_new))
            opx.cost = np.concatenate((opx.cost, zeros_to_new))
            opx.vat_portion = np.concatenate((opx.vat_portion, zeros_to_new))
            opx.vat_discount = np.concatenate((opx.vat_discount, zeros_to_new))
            opx.lbt_portion = np.concatenate((opx.lbt_portion, zeros_to_new))
            opx.lbt_discount = np.concatenate((opx.lbt_discount, zeros_to_new))

            opx.project_duration = project_duration_trans
            opx.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract1 asr
        for asr in self.contract1.asr_cost:
            asr.start_year = start_date_trans.year
            asr.end_year = end_date_trans.year

            asr.cost = np.concatenate((asr.cost, zeros_to_new))
            asr.expense_year = np.concatenate((asr.expense_year, years_to_new)).astype(int)
            asr.cost_allocation = asr.cost_allocation + fluid_to_new
            asr.description = asr.description + desc_to_new
            asr.vat_portion = np.concatenate((asr.vat_portion, zeros_to_new))
            asr.vat_discount = np.concatenate((asr.vat_discount, zeros_to_new))
            asr.lbt_portion = np.concatenate((asr.lbt_portion, zeros_to_new))
            asr.lbt_discount = np.concatenate((asr.lbt_discount, zeros_to_new))

            asr.project_duration = project_duration_trans
            asr.project_years = project_years_trans

        # Modifying the contract2
        # Changing the start_date and end_date of contract2
        self.contract2.start_date = start_date_trans
        self.contract2.end_date = end_date_trans
        self.contract2.project_duration = project_duration_trans
        self.contract2.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract2 lifting
        for lift in self.contract2.lifting:
            lift.start_year = start_date_trans.year
            lift.end_year = end_date_trans.year

            lift.lifting_rate = np.concatenate((zeros_to_prior, lift.lifting_rate))
            lift.price = np.concatenate((zeros_to_prior, lift.price))
            lift.ghv = np.concatenate((zeros_to_prior, lift.ghv))
            lift.prod_rate = np.concatenate((zeros_to_prior, lift.prod_rate))
            lift.prod_year = np.concatenate((years_to_prior, lift.prod_year)).astype(int)

            lift.project_duration = project_duration_trans
            lift.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract2 tangible
        for tang in self.contract2.tangible_cost:
            tang.start_year = start_date_trans.year
            tang.end_year = end_date_trans.year

            tang.cost = np.concatenate((zeros_to_prior, tang.cost))
            tang.expense_year = np.concatenate((years_to_prior, tang.expense_year)).astype(int)
            tang.cost_allocation = fluid_to_prior + tang.cost_allocation
            tang.description = desc_to_prior + tang.description
            tang.vat_portion = np.concatenate((zeros_to_prior, tang.vat_portion))
            tang.vat_discount = np.concatenate((zeros_to_prior, tang.vat_discount))
            tang.lbt_portion = np.concatenate((zeros_to_prior, tang.lbt_portion))
            tang.lbt_discount = np.concatenate((zeros_to_prior, tang.lbt_discount))
            tang.pis_year = np.concatenate((zeros_to_prior, tang.pis_year)).astype(int)
            tang.salvage_value = np.concatenate((zeros_to_prior, tang.salvage_value))
            tang.useful_life = np.concatenate((zeros_to_prior, tang.useful_life))
            tang.depreciation_factor = np.concatenate((zeros_to_prior, tang.depreciation_factor))
            tang.is_ic_applied = ic_to_prior + tang.is_ic_applied

            tang.project_duration = project_duration_trans
            tang.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract2 intangible
        for intang in self.contract2.intangible_cost:
            intang.start_year = start_date_trans.year
            intang.end_year = end_date_trans.year

            intang.cost = np.concatenate((zeros_to_prior, intang.cost))
            intang.expense_year = np.concatenate((years_to_prior, intang.expense_year)).astype(int)
            intang.cost_allocation = fluid_to_prior + intang.cost_allocation
            intang.description = desc_to_prior + intang.description
            intang.vat_portion = np.concatenate((zeros_to_prior, intang.vat_portion))
            intang.vat_discount = np.concatenate((zeros_to_prior, intang.vat_discount))
            intang.lbt_portion = np.concatenate((zeros_to_prior, intang.lbt_portion))
            intang.lbt_discount = np.concatenate((zeros_to_prior, intang.lbt_discount))

            intang.project_duration = project_duration_trans
            intang.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract2 opex
        for opx in self.contract2.opex:
            opx.start_year = start_date_trans.year
            opx.end_year = end_date_trans.year

            opx.expense_year = np.concatenate((years_to_prior, opx.expense_year)).astype(int)
            opx.cost_allocation = fluid_to_prior + opx.cost_allocation
            opx.description = desc_to_prior + opx.description
            opx.fixed_cost = np.concatenate((zeros_to_prior, opx.fixed_cost))
            opx.prod_rate = np.concatenate((zeros_to_prior, opx.prod_rate))
            opx.cost_per_volume = np.concatenate((zeros_to_prior, opx.cost_per_volume))
            opx.variable_cost = np.concatenate((zeros_to_prior, opx.variable_cost))
            opx.cost = np.concatenate((zeros_to_prior, opx.cost))
            opx.vat_portion = np.concatenate((zeros_to_prior, opx.vat_portion))
            opx.vat_discount = np.concatenate((zeros_to_prior, opx.vat_discount))
            opx.lbt_portion = np.concatenate((zeros_to_prior, opx.lbt_portion))
            opx.lbt_discount = np.concatenate((zeros_to_prior, opx.lbt_discount))

            opx.project_duration = project_duration_trans
            opx.project_years = project_years_trans

        # Concatenating the zeros_to_new and years_to_new to the contract2 asr
        for asr in self.contract2.asr_cost:
            asr.start_year = start_date_trans.year
            asr.end_year = end_date_trans.year

            asr.cost = np.concatenate((zeros_to_prior, asr.cost))
            asr.expense_year = np.concatenate((years_to_prior, asr.expense_year)).astype(int)
            asr.cost_allocation = fluid_to_prior + asr.cost_allocation
            asr.description = desc_to_prior + asr.description
            asr.vat_portion = np.concatenate((zeros_to_prior, asr.vat_portion))
            asr.vat_discount = np.concatenate((zeros_to_prior, asr.vat_discount))
            asr.lbt_portion = np.concatenate((zeros_to_prior, asr.lbt_portion))
            asr.lbt_discount = np.concatenate((zeros_to_prior, asr.lbt_discount))

            asr.project_duration = project_duration_trans
            asr.project_years = project_years_trans

        # Parsing the attributes to the new object of contract
        contract1_new = self._parse_dataclass(contract=self.contract1)
        contract2_new = self._parse_dataclass(contract=self.contract2)

        # Executing the new contract
        contract1_new.run(**self.argument_contract1)
        contract2_new.run(**self.argument_contract2)

        # Defining the unrecoverable cost from the prior contract
        # Condition if the contract is Cost Recovery
        if isinstance(contract1_new, CostRecovery):
            latest_unrec = contract1_new._consolidated_unrecovered_after_transfer[-1]
        # Condition if the contract is Gross Split
        else:
            latest_unrec = contract1_new._consolidated_carward_cost_aftertf[-1]
        unrec_cost_prior = np.zeros_like(contract2_new.project_years, dtype=float)
        unrec_cost_prior[len(years_to_prior)] = latest_unrec

        # Calculating the Unrecoverable cost in the transitioned cashflow
        unrec_trans = psc_tools.get_unrecovered_cost(revenue=contract2_new._consolidated_taxable_income,
                                                     depreciation=unrec_cost_prior,
                                                     non_capital=np.zeros_like(contract2_new.project_years,
                                                                               dtype=float),
                                                     ftp_ctr=np.zeros_like(contract2_new.project_years, dtype=float),
                                                     ftp_gov=np.zeros_like(contract2_new.project_years, dtype=float),
                                                     ic=np.zeros_like(contract2_new.project_years, dtype=float))

        cost_to_be_recovered_trans = psc_tools.get_cost_to_be_recovered(unrecovered_cost=unrec_trans)

        cost_recovery_trans = CostRecovery._get_cost_recovery(
            revenue=contract2_new._consolidated_taxable_income,
            ftp=np.zeros_like(contract2_new.project_years, dtype=float),
            ic=np.zeros_like(contract2_new.project_years, dtype=float),
            depreciation=unrec_cost_prior,
            non_capital=np.zeros_like(contract2_new.project_years, dtype=float),
            cost_to_be_recovered=cost_to_be_recovered_trans,
            cr_cap_rate=1.0)

        # Calculate taxable income after considering transferred unrecoverable cost from contract1
        taxable_income_trans = CostRecovery._get_ets_before_transfer(revenue=contract2_new._consolidated_taxable_income,
                                                                     ftp_ctr=np.zeros_like(contract2_new.project_years,
                                                                                           dtype=float),
                                                                     ftp_gov=np.zeros_like(contract2_new.project_years,
                                                                                           dtype=float),
                                                                     ic=np.zeros_like(contract2_new.project_years,
                                                                                      dtype=float),
                                                                     cost_recovery=cost_recovery_trans)

        # Calculate tax payment of adjusted taxable income
        tax_payment_transition = taxable_income_trans * contract2_new._tax_rate_arr

        # Calculate new ctr net share
        contract2_new._consolidated_ctr_net_share = (contract2_new._consolidated_taxable_income -
                                                     (unrec_cost_prior * unrec_portion) -
                                                     tax_payment_transition)

        # Calculate new cashflow of contract2
        contract2_new._consolidated_cashflow = (contract2_new._consolidated_cashflow +
                                                contract2_new._consolidated_tax_payment -
                                                tax_payment_transition
                                                )

        # Calculate new government take
        contract2_new._consolidated_government_take = (contract2_new._consolidated_government_take -
                                                       contract2_new._consolidated_tax_payment +
                                                       tax_payment_transition)

        contract2_new._consolidated_tax_payment = tax_payment_transition

        # Parsing new contract to the attributes of Transition dataclass
        self._contract1_transitioned = contract1_new
        self._contract2_transitioned = contract2_new

        # Populating transition attributes
        # Lifting
        self._oil_lifting = (self._contract1_transitioned._oil_lifting +
                             self._contract2_transitioned._oil_lifting)

        self._gas_lifting = (self._contract1_transitioned._gas_lifting +
                             self._contract2_transitioned._gas_lifting)

        self._sulfur_lifting = (self._contract1_transitioned._sulfur_lifting +
                                self._contract2_transitioned._sulfur_lifting)

        self._electricity_lifting = (self._contract1_transitioned._electricity_lifting +
                                     self._contract2_transitioned._electricity_lifting)

        self._co2_lifting = (self._contract1_transitioned._co2_lifting +
                             self._contract2_transitioned._co2_lifting)

        # Revenue
        self._oil_revenue = (self._contract1_transitioned._oil_revenue +
                             self._contract2_transitioned._oil_revenue)

        self._gas_revenue = (self._contract1_transitioned._gas_revenue +
                             self._contract2_transitioned._gas_revenue)

        self._sulfur_revenue = (self._contract1_transitioned._sulfur_revenue +
                                self._contract2_transitioned._sulfur_revenue)

        self._electricity_revenue = (self._contract1_transitioned._electricity_revenue +
                                     self._contract2_transitioned._electricity_revenue)

        self._co2_revenue = (self._contract1_transitioned._co2_revenue +
                             self._contract2_transitioned._co2_revenue)

        # Tangible
        self._oil_tangible_expenditures = (self._contract1_transitioned._oil_tangible_expenditures +
                                           self._contract2_transitioned._oil_tangible_expenditures)

        self._gas_tangible_expenditures = (self._contract1_transitioned._gas_tangible_expenditures +
                                           self._contract2_transitioned._gas_tangible_expenditures)

        # Intangible
        self._oil_intangible_expenditures = (self._contract1_transitioned._oil_intangible_expenditures +
                                             self._contract2_transitioned._oil_intangible_expenditures)

        self._gas_intangible_expenditures = (self._contract1_transitioned._gas_intangible_expenditures +
                                             self._contract2_transitioned._gas_intangible_expenditures)

        # Opex
        self._oil_opex_expenditures = (self._contract1_transitioned._oil_opex_expenditures +
                                       self._contract2_transitioned._oil_opex_expenditures)

        self._gas_opex_expenditures = (self._contract1_transitioned._gas_opex_expenditures +
                                       self._contract2_transitioned._gas_opex_expenditures)

        # ASR
        self._oil_asr_expenditures = (self._contract1_transitioned._oil_asr_expenditures +
                                      self._contract2_transitioned._oil_asr_expenditures)

        self._gas_asr_expenditures = (self._contract1_transitioned._gas_asr_expenditures +
                                      self._contract2_transitioned._gas_asr_expenditures)

        # WAP Price
        self._oil_wap_price = self._contract2_transitioned._oil_wap_price
        self._gas_wap_price = self._contract2_transitioned._gas_wap_price
        self._sulfur_wap_price = self._contract2_transitioned._sulfur_wap_price
        self._electricity_wap_price = self._contract2_transitioned._electricity_wap_price
        self._co2_wap_price = self._contract2_transitioned._co2_wap_price

        # Sunk Cost
        self._oil_sunk_cost = self._contract1_transitioned._oil_sunk_cost + self._contract2_transitioned._oil_sunk_cost
        self._gas_sunk_cost = self._contract1_transitioned._gas_sunk_cost + self._contract2_transitioned._gas_sunk_cost

        # Deductible Cost / Cost Recovery
        _oil_deductible_cost_1 = None
        _gas_deductible_cost_1 = None
        _oil_deductible_cost_2 = None
        _gas_deductible_cost_2 = None

        _oil_unrec_cost_1 = None
        _gas_unrec_cost_1 = None
        _oil_unrec_cost_2 = None
        _gas_unrec_cost_2 = None

        _oil_ctr_ets_1 = None
        _gas_ctr_ets_1 = None
        _oil_ctr_ets_2 = None
        _gas_ctr_ets_2 = None

        _oil_gov_ets_1 = None
        _gas_gov_ets_1 = None
        _oil_gov_ets_2 = None
        _gas_gov_ets_2 = None

        _ctr_ftp_1 = None
        _gov_ftp_1 = None
        _ctr_ftp_2 = None
        _gov_ftp_2 = None

        # Determining variables if the type of Contract1 is Cost Recovery
        if isinstance(self.contract1, CostRecovery):
            _oil_deductible_cost_1 = self._contract1_transitioned._oil_cost_recovery_after_tf
            _gas_deductible_cost_1 = self._contract1_transitioned._gas_cost_recovery_after_tf
            _oil_unrec_cost_1 = self._contract1_transitioned._oil_unrecovered_after_transfer
            _gas_unrec_cost_1 = self._contract1_transitioned._gas_unrecovered_after_transfer
            _oil_ctr_ets_1 = self._contract1_transitioned._oil_ets_after_transfer
            _gas_ctr_ets_1 = self._contract1_transitioned._gas_ets_after_transfer
            _oil_gov_ets_1 = self._contract1_transitioned._oil_government_share
            _gas_gov_ets_1 = self._contract1_transitioned._gas_government_share
            _ctr_ftp_1 = self._contract1_transitioned._consolidated_ftp_ctr
            _gov_ftp_1 = self._contract1_transitioned._consolidated_ftp_gov

        # Determining variables if the type of Contract1 is Gross Split
        elif isinstance(self.contract1, GrossSplit):
            _oil_deductible_cost_1 = self._contract1_transitioned._oil_deductible_cost
            _gas_deductible_cost_1 = self._contract1_transitioned._gas_deductible_cost
            _oil_unrec_cost_1 = self._contract1_transitioned._oil_carward_deduct_cost
            _gas_unrec_cost_1 = self._contract1_transitioned._gas_carward_deduct_cost
            _oil_ctr_ets_1 = self._contract1_transitioned._oil_ctr_share_after_transfer
            _gas_ctr_ets_1 = self._contract1_transitioned._gas_ctr_share_after_transfer
            _oil_gov_ets_1 = np.zeros_like(self._contract1_transitioned.project_years)
            _gas_gov_ets_1 = np.zeros_like(self._contract1_transitioned.project_years)
            _ctr_ftp_1 = np.zeros_like(self._contract1_transitioned.project_years, dtype=float)
            _gov_ftp_1 = self._contract1_transitioned._consolidated_gov_share_before_tf

        # Determining variables if the type of Contract2 is Cost Recovery
        if isinstance(self.contract2, CostRecovery):
            _oil_deductible_cost_2 = self._contract2_transitioned._oil_cost_recovery_after_tf
            _gas_deductible_cost_2 = self._contract2_transitioned._gas_cost_recovery_after_tf
            _oil_unrec_cost_2 = self._contract2_transitioned._oil_unrecovered_after_transfer
            _gas_unrec_cost_2 = self._contract2_transitioned._gas_unrecovered_after_transfer
            _oil_ctr_ets_2 = self._contract2_transitioned._oil_ets_after_transfer
            _gas_ctr_ets_2 = self._contract2_transitioned._gas_ets_after_transfer
            _oil_gov_ets_2 = self._contract2_transitioned._oil_government_share
            _gas_gov_ets_2 = self._contract2_transitioned._gas_government_share
            _ctr_ftp_2 = self._contract2_transitioned._consolidated_ftp_ctr
            _gov_ftp_2 = self._contract2_transitioned._consolidated_ftp_gov

        # Determining variables if the type of Contract2 is Gross Split
        elif isinstance(self.contract2, GrossSplit):
            _oil_deductible_cost_2 = self._contract2_transitioned._oil_deductible_cost
            _gas_deductible_cost_2 = self._contract2_transitioned._gas_deductible_cost
            _oil_unrec_cost_2 = self._contract2_transitioned._oil_carward_deduct_cost
            _gas_unrec_cost_2 = self._contract2_transitioned._gas_carward_deduct_cost
            _oil_ctr_ets_2 = self._contract2_transitioned._oil_ctr_share_after_transfer
            _gas_ctr_ets_2 = self._contract2_transitioned._gas_ctr_share_after_transfer
            _oil_gov_ets_2 = np.zeros_like(self._contract2_transitioned.project_years)
            _gas_gov_ets_2 = np.zeros_like(self._contract2_transitioned.project_years)
            _ctr_ftp_2 = np.zeros_like(self._contract2_transitioned.project_years, dtype=float)
            _gov_ftp_2 = self._contract2_transitioned._consolidated_gov_share_before_tf

        # Deductible Cost / Cost Recovery
        self._oil_deductible_cost = _oil_deductible_cost_1 + _oil_deductible_cost_2
        self._gas_deductible_cost = _gas_deductible_cost_1 + _gas_deductible_cost_2

        # Carry Forward Cost / Unrecoverable Cost
        self._oil_unrec_cost = _oil_unrec_cost_2
        self._gas_unrec_cost = _gas_unrec_cost_2

        # ETS
        self._oil_ctr_ets = _oil_ctr_ets_1 + _oil_ctr_ets_2
        self._gas_ctr_ets = _gas_ctr_ets_1 + _gas_ctr_ets_2

        self._oil_gov_ets = _oil_gov_ets_1 + _oil_gov_ets_2
        self._gas_gov_ets = _gas_gov_ets_1 + _gas_gov_ets_2

        # Contractor Net Operating Profit
        self._net_operating_profit = (self._contract1_transitioned._consolidated_ctr_net_share +
                                      self._contract2_transitioned._consolidated_ctr_net_share)

        # Contractor Net Cash Flow
        self._ctr_net_cashflow = (self._contract1_transitioned._consolidated_cashflow +
                                  self._contract2_transitioned._consolidated_cashflow)

        # FTP
        self._ctr_ftp = _ctr_ftp_1 + _ctr_ftp_2
        self._gov_ftp = _gov_ftp_1 + _gov_ftp_2

        # DDMO
        self._ddmo = (self._contract1_transitioned._consolidated_ddmo +
                      self._contract2_transitioned._consolidated_ddmo)

        # Tax
        self._tax_payment = (self._contract1_transitioned._consolidated_tax_payment +
                             self._contract2_transitioned._consolidated_tax_payment)

        # Government Take
        self._government_take = (self._contract1_transitioned._consolidated_government_take +
                                 self._contract2_transitioned._consolidated_government_take)

        # Consolidated Attributes
        # Consolidated Revenue
        self._consolidated_revenue = self._oil_revenue + self._gas_revenue

        # Consolidated Sunk Cost
        self._consolidated_sunk_cost = self._oil_sunk_cost + self._gas_sunk_cost

        # Consolidated Cash Flow
        self._consolidated_cashflow = (self._contract1_transitioned._consolidated_cashflow +
                                       self._contract2_transitioned._consolidated_cashflow)

        # Consolidated DDMO
        self._consolidated_ddmo = (self._contract1_transitioned._consolidated_ddmo +
                                   self._contract2_transitioned._consolidated_ddmo)

        # Consolidated Tax Payment
        self._consolidated_tax_payment = (self._contract1_transitioned._consolidated_tax_payment +
                                          self._contract2_transitioned._consolidated_tax_payment)

        # Consolidated Government Take
        self._consolidated_government_take = (self._contract1_transitioned._consolidated_government_take +
                                              self._contract2_transitioned._consolidated_government_take)

        # Project Years
        self.project_years = np.copy(self._contract1_transitioned.project_years)























"""
Specify constants for several processes using Enum class.
"""

from enum import Enum


class DeprMethod(Enum):
    """
    Enumeration of depreciation methods.

    Attributes
    ----------
    SL : str
        Represents the straight-line depreciation method
    DB : str
        Represents the declining balance depreciation method
    PSC_DB : str
        Represents the PSC double declining balance method
    UOP : str
        Represents the units of production depreciation method
    """

    SL = "sl"
    DB = "db"
    PSC_DB = "psc db"
    UOP = "uop"


class FluidType(Enum):
    """
    Enumeration of fluid types for depreciation calculation.

    Attributes
    ----------
    ALL : str
        Represents all fluid types
    OIL : str
        Represents oil as the fluid type
    GAS : str
        Represents gas as the fluid type
    SULFUR: str
        Represents sulfur as the fluid type
    ELECTRICITY: str
        Represent electricity as the fluid type
    CO2: str
        Represents CO2 as the fluid type
    """

    ALL = "all"
    OIL = "oil"
    GAS = "gas"
    SULFUR = "sulfur"
    ELECTRICITY = "electricity"
    CO2 = "co2"


class TaxType(Enum):
    """
    Enumeration of tax types for incorporation of tax adjustment.

    Attributes
    ----------
    VAT : str
        Represents Value Added Tax.
    LBT : str
        Represents Land and Building Tax.
    """

    VAT = "vat"
    LBT = "lbt"


class TaxSplitTypeCR(Enum):
    """
    Enumeration of split type for cost recovery.

    Attributes
    ----------
    CONVENTIONAL : str
        Represents conventional split type
    SLIDING_SCALE : str
        Represents sliding scale split type
    R2C : str
        Represents revenue to cost split type
    """

    CONVENTIONAL = "conventional"
    SLIDING_SCALE = "ICP sliding scale"
    R2C = "R/C"


class YearReference(Enum):
    """
    Enumeration of different year references used for depreciation calculations.

    Attributes:
    -----------
    PIS_YEAR : str
        Symbolic name for the Placed in Service (PIS) year reference

    EXPENSE_YEAR : str
        Symbolic name for the expense year reference
    """

    PIS_YEAR = "pis_year"
    EXPENSE_YEAR = "expense_year"


class TaxRegime(Enum):
    """
    Enumeration class representing different tax regimes for oil and gas economic analysis.

    Attributes
    ----------
    UU_36_2008 (str): Tax regime under "UU No.36 Tahun 2008"
    UU_02_2020 (str): Tax regime under "UU No.02 Tahun 2020"
    UU_07_2021 (str): Tax regime under "UU No.07 Tahun 2021"
    NAILED_DOWN (str): Special tax regime designated as "nailed down"
    PREVAILING (str): General tax regime referred to as "prevailing"
    """
    NAILED_DOWN = "nailed down"
    PREVAILING = "prevailing"
    UU_36_2008 = "UU No.36 Tahun 2008"
    UU_02_2020 = "UU No.02 Tahun 2020"
    UU_07_2021 = "UU No.07 Tahun 2021"


class TaxPaymentMode(Enum):
    """
    Enumeration class representing different tax payment modes.

    Attributes
    ----------
    TAX_DUE_MODE: str
        Calculation method where it is treated like an unrecoverable cost,
        and could be weighted on the next year.
    TAX_DIRECT_MODE: str
        Calculation method where the tax is calculated directly from taxable income.
    """
    TAX_DUE_MODE = (
        "Calculation method where it treated like the unrecoverable cost, "
        "could be weighted on the next year"
    )
    TAX_DIRECT_MODE = (
        "Calculation method where the tax calculated directly from taxable income"
    )


class FTPTaxRegime(Enum):
    """
    An enumeration depicting the FTP Tax Regime associated with oil and gas
    Production Sharing Contract in Indonesia.

    Attributes
    ----------
    PDJP_20_2017: str
        Manifests Peraturan DIRJEN PAJAK No. PER-20/PJ/2017
    PRE_PDJP_20_2017: str
        Manifests Pre Peraturan DIRJEN PAJAK No. PER-20/PJ/2017
    """
    PDJP_20_2017 = "Peraturan DIRJEN PAJAK No. PER-20/PJ/2017"
    PRE_PDJP_20_2017 = "Pre Peraturan DIRJEN PAJAK No. PER-20/PJ/2017"
    DIRECT_MODE = "The tax will be paid directly"


class GrossSplitRegime(Enum):
    """
    An enumeration representing different Gross Split Regimes in the context of
    oil and gas economic analysis, specifically related to Production Sharing Contracts
    in Indonesia.

    Attributes
    ----------
    PERMEN_ESDM_8_2017: str
        Represents the Gross Split Regime according to Peraturan Menteri ESDM No. 8 Tahun 2017.
    PERMEN_ESDM_52_2017: str
        Represents the Gross Split Regime according to Peraturan Menteri ESDM No. 52 Tahun 2017.
    PERMEN_ESDM_20_2019: str
        Represents the Gross Split Regime according to Peraturan Menteri ESDM No. 20 Tahun 2019.
    PERMEN_ESDM_12_2020: str
        Represents the Gross Split Regime according to Peraturan Menteri ESDM No. 12 Tahun 2020.
    """
    PERMEN_ESDM_8_2017 = 'Peraturan Menteri ESDM No. 8 Tahun 2017'
    PERMEN_ESDM_52_2017 = 'Peraturan Menteri ESDM No. 52 Tahun 2017'
    PERMEN_ESDM_20_2019 = 'Peraturan Menteri ESDM No. 20 Tahun 2019'
    PERMEN_ESDM_12_2020 = 'Peraturan Menteri ESDM No. 12 Tahun 2020'


class DiscountingMode(Enum):
    """
    Enumeration class representing different discounting modes for financial calculations.

    Attributes
    ----------
    END_YEAR: str
        Discounting method where the cash flows are discounted to the end of each year.
    MID_YEAR: str
        Discounting method where the cash flows are discounted to the middle of each year.
    """
    END_YEAR = "End Year Discounting"
    MID_YEAR = "Mid Year Discounting"


class ContractType(Enum):
    """
    Enumeration class representing different types of production sharing contracts
    (PSCs) in Indonesia.

    Attributes
    ----------
    COST_RECOVERY: str
        Cost Recovery (CR) contract type.
    GROSS_SPLIT: str
        Gross Split (GS) contract type.
    TRANSITION_CR_CR: str
        Transition contract type from Cost Recovery to Cost Recovery.
    TRANSITION_CR_GS: str
        Transition contract type from Cost Recovery to Gross Split.
    TRANSITION_GS_GS: str
        Transition contract type from Gross Split to Gross Split.
    TRANSITION_GS_CR: str
        Transition contract type from Gross Split to Cost Recovery.
    """
    COST_RECOVERY = 'Cost Recovery (CR)'
    GROSS_SPLIT = 'Gross Split (GS)'
    TRANSITION_CR_CR = 'Transition CR - CR'
    TRANSITION_CR_GS = 'Transition CR - GS'
    TRANSITION_GS_GS = 'Transition GS - GS'
    TRANSITION_GS_CR = 'Transition GS - CR'


class NPVSelection(Enum):
    """
    Enumeration class representing different methods for calculating
    Net Present Value (NPV) in economic analysis.

    Attributes
    ----------
    NPV_REAL_TERMS: str
        NPV Calculation using the real terms method.
    NPV_NOMINAL_TERMS: str
        NPV Calculation using the nominal terms method.
    NPV_SKK_NOMINAL_TERMS: str
        NPV Calculation using the SKK nominal terms method.
    NPV_SKK_REAL_TERMS: str
        NPV Calculation using the SKK real terms method.
    NPV_POINT_FORWARD: str
        NPV Calculation using the Point Forward method.
    """
    NPV_REAL_TERMS = 'NPV Calculation using real terms method'
    NPV_NOMINAL_TERMS = 'NPV Calculation using nominal terms method'
    NPV_SKK_NOMINAL_TERMS = 'NPV Calculation using SKK Nominal terms method'
    NPV_SKK_REAL_TERMS = 'NPV Calculation using SKK Real terms method'
    NPV_POINT_FORWARD = 'NPV Calculation using Point Forward method'


class OtherRevenue(Enum):
    """
    Enumeration class representing different types of other revenue
    in oil and gas economic analysis.

    Attributes
    ----------
    ADDITION_TO_OIL_REVENUE: str
        Revenue treated as addition to oil revenue.
    ADDITION_TO_GAS_REVENUE: str
        Revenue treated as addition to gas revenue.
    REDUCTION_TO_OIL_OPEX: str
        Revenue treated as reduction to oil operational expenditure (OPEX).
    REDUCTION_TO_GAS_OPEX: str
        Revenue treated as reduction to gas operational expenditure (OPEX).
    """
    ADDITION_TO_OIL_REVENUE = "The revenue will be treated as the Oil Revenue Addition"
    ADDITION_TO_GAS_REVENUE = "The revenue will be treated as the Gas Revenue Addition"
    REDUCTION_TO_OIL_OPEX = "The revenue will be treated as the Oil OPEX Reduction"
    REDUCTION_TO_GAS_OPEX = "The revenue will be treated as the GAS OPEX Reduction"


class InflationAppliedTo(Enum):
    """
    Enumeration class representing different aspects to which inflation rate
    is applied in economic analysis.

    Attributes
    ----------
    CAPEX: str
        Inflation rate applied to Tangible and Intangible Costs (CAPEX).
    OPEX: str
        Inflation rate applied to Operating Costs (OPEX).
    CAPEX_AND_OPEX: str
        Inflation rate applied to Tangible, Intangible, and Operating Costs (CAPEX and OPEX).
    """
    CAPEX = "Inflation rate will be applied to Tangible and Intangible Cost"
    OPEX = "Inflation rate will be applied to Operating Cost"
    CAPEX_AND_OPEX = "Inflation rate will be applied to Tangible, Intangible, Operating Cost"


class OptimizationParameter(Enum):
    """
    Enumeration class representing different optimization parameters in
    oil and gas economic analysis.

    Attributes
    ----------
    OIL_CTR_PRETAX: str
        Optimization parameter for Oil Contractor Pre Tax.
    GAS_CTR_PRETAX: str
        Optimization parameter for Gas Contractor Pre Tax.
    OIL_FTP_PORTION: str
        Optimization parameter for Oil FTP Portion.
    GAS_FTP_PORTION: str
        Optimization parameter for Gas FTP Portion.
    OIL_IC: str
        Optimization parameter for Oil Investment Credit (IC).
    GAS_IC: str
        Optimization parameter for Gas Investment Credit (IC).
    OIL_DMO_FEE: str
        Optimization parameter for Oil Domestic Market Obligation (DMO) Fee.
    GAS_DMO_FEE: str
        Optimization parameter for Gas Domestic Market Obligation (DMO) Fee.
    VAT_RATE: str
        Optimization parameter for Value-Added Tax (VAT) Rate.
    EFFECTIVE_TAX_RATE: str
        Optimization parameter for Effective Tax Rate.
    MINISTERIAL_DISCRETION: str
        Optimization parameter for Ministerial Discretion.
    """
    OIL_CTR_PRETAX = 'Oil Contractor Pre Tax'
    GAS_CTR_PRETAX = 'Gas Contractor Pre Tax'
    OIL_FTP_PORTION = 'Oil FTP Portion'
    GAS_FTP_PORTION = 'Gas FTP Portion'
    OIL_IC = 'Oil IC'
    GAS_IC = 'Gas IC'
    OIL_DMO_FEE = 'Oil DMO Fee'
    GAS_DMO_FEE = 'Gas Dmo Fee'
    VAT_RATE = 'VAT Rate'
    EFFECTIVE_TAX_RATE = 'Effective Tax Rate'
    MINISTERIAL_DISCRETION = 'Ministerial Discretion'


class OptimizationTarget(Enum):
    """
    Enumeration class representing different optimization targets in economic analysis.

    Attributes
    ----------
    IRR: str
        Optimization target for Internal Rate of Return (IRR).
    NPV: str
        Optimization target for Net Present Value (NPV).
    PI: str
        Optimization target for Profitability Index (PI).
    """
    IRR = 'IRR'
    NPV = 'NPV'
    PI = 'PI'
